@@ -477,7 +477,7 @@ void __connman_ipconfig_newlink(int index, unsigned short type,
 						struct rtnl_link_stats *stats)
 {
 	struct connman_ipdevice *ipdevice;
-	GList *list;
+	GList *list, *ipconfig_copy;
 	GString *str;
 	bool up = false, down = false;
 	bool lower_up = false, lower_down = false;
@@ -556,7 +556,9 @@ update:
 
 	g_string_free(str, TRUE);
 
-	for (list = g_list_first(ipconfig_list); list;
+	ipconfig_copy = g_list_copy(ipconfig_list);
+
+	for (list = g_list_first(ipconfig_copy); list;
 						list = g_list_next(list)) {
 		struct connman_ipconfig *ipconfig = list->data;
 
@@ -577,6 +579,8 @@ update:
 			ipconfig->ops->down(ipconfig, ifname);
 	}
 
+	g_list_free(ipconfig_copy);
+
 	if (lower_up)
 		__connman_ipconfig_lower_up(ipdevice);
 	if (lower_down)
@@ -1086,8 +1090,6 @@ int __connman_ipconfig_gateway_add(struct connman_ipconfig *ipconfig)
 	if (!service)
 		return -EINVAL;
 
-	__connman_connection_gateway_remove(service, ipconfig->type);
-
 	DBG("type %d gw %s peer %s", ipconfig->type,
 		ipconfig->address->gateway, ipconfig->address->peer);
 
