@@ -69,27 +69,13 @@ struct vpn_ipdevice {
 
 static GHashTable *ipdevice_hash = NULL;
 
-unsigned char __vpn_ipconfig_netmask_prefix_len(const char *netmask)
+struct connman_ipaddress *
+__vpn_ipconfig_get_address(struct vpn_ipconfig *ipconfig)
 {
-	unsigned char bits;
-	in_addr_t mask;
-	in_addr_t host;
+	if (!ipconfig)
+		return NULL;
 
-	if (!netmask)
-		return 32;
-
-	mask = inet_network(netmask);
-	host = ~mask;
-
-	/* a valid netmask must be 2^n - 1 */
-	if ((host & (host + 1)) != 0)
-		return -1;
-
-	bits = 0;
-	for (; mask; mask <<= 1)
-		++bits;
-
-	return bits;
+	return ipconfig->address;
 }
 
 const char *__vpn_ipconfig_get_peer(struct vpn_ipconfig *ipconfig)
